/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.persistence.impl;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.annotation.Annotation;
import java.lang.ref.WeakReference;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.concurrent.ConcurrentHashMap;

import javax.annotation.PostConstruct;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Service;
import org.springframework.util.FileCopyUtils;
import org.springframework.util.FileSystemUtils;
import org.springframework.util.WeakReferenceMonitor;
import org.springframework.util.WeakReferenceMonitor.ReleaseListener;

import com.google.common.io.Files;

import kieker.analysis.AnalysisController;
import kieker.analysis.model.analysisMetaModel.MIAnalysisComponent;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIFilter;
import kieker.analysis.model.analysisMetaModel.MIInputPort;
import kieker.analysis.model.analysisMetaModel.MIOutputPort;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIProperty;
import kieker.analysis.model.analysisMetaModel.MIRepository;
import kieker.analysis.model.analysisMetaModel.MIRepositoryConnector;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.repository.AbstractRepository;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.ClassAndMethodContainer;
import kieker.webgui.common.exception.ComponentInitializationException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;
import kieker.webgui.common.exception.ProjectNotExistingException;
import kieker.webgui.domain.ComponentListContainer;
import kieker.webgui.domain.PluginContainer;
import kieker.webgui.domain.RepositoryContainer;
import kieker.webgui.persistence.IProjectDAO;
import kieker.webgui.persistence.impl.util.CloseableURLClassLoader;
import kieker.webgui.persistence.impl.util.PluginFinder;

import org.primefaces.model.UploadedFile;

/**
 * This is an implementation of the {@link IProjectDAO} interface, which uses the file system to store the available projects and everything.
 * 
 * @author Nils Christian Ehmke
 */
@Service
public class FSProjectDAOImpl implements IProjectDAO, ReleaseListener {

	private static final Log LOG = LogFactory.getLog(FSProjectDAOImpl.class);

	private static final MIAnalysisMetaModelFactory FACTORY = MIAnalysisMetaModelFactory.eINSTANCE;
	private static final String KIEKER_LIB = "kieker-1.7-SNAPSHOT_emf.jar";
	private static final String META_FILE = "meta.dat";
	private static final String KAX_EXTENSION = "kax";
	private static final String LIB_EXTENSION = "jar";
	private static final String LIB_DIRECTORY = "lib";
	private static final String ROOT_DIRECTORY = "data";

	@Autowired
	private PluginFinder pluginFinder;

	private final Map<CloseableURLClassLoader, WeakReference<Object>> classLoaders = new ConcurrentHashMap<CloseableURLClassLoader, WeakReference<Object>>();
	private final Map<File, WeakReference<Object>> tempDirs = new ConcurrentHashMap<File, WeakReference<Object>>();
	private final Map<String, ComponentListContainer> availableComponents = new ConcurrentHashMap<String, ComponentListContainer>();

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public FSProjectDAOImpl() {
		// no code necessary
	}

	/**
	 * This method initializes the object. <b>Do not call this method manually. It will only be accessed by Spring.</b>
	 * 
	 * @throws IOException
	 *             If either the creation of the necessary directories failed or if something went wrong during the initialization of the projects.
	 */
	@PostConstruct
	public void initialize() throws IOException {
		// Check for the necessary directories and create them if necessary
		this.checkDir(FSProjectDAOImpl.ROOT_DIRECTORY);
		// This method call will initialize the available components for all projects which currently exist.
		this.initializeAvailableComponents();
	}

	/**
	 * This method initializes the available components for all projects. This means that this method will run through all projects and load all available plugins,
	 * filters and repositories from the existing libraries - including Kieker.
	 * 
	 * @throws IOException
	 *             If something went really wrong during the initialization of one of the projects.
	 */
	private void initializeAvailableComponents() throws IOException {
		final Collection<String> projects = this.listAllProjects();
		// Initialize all available components within all projects
		for (final String project : projects) {
			this.initializeAvailableComponents(project);
		}
	}

	/**
	 * This method initializes the available components for the given project. This means that this method will load all available plugins, filters and repositories
	 * from the existing libraries - including Kieker.
	 * 
	 * @param project
	 *            The name of the project to be initialized.
	 * @throws IOException
	 *             If something went really wrong during the initialization of the project.
	 */
	private void initializeAvailableComponents(final String project) throws IOException {
		try {
			final Object dummyObject = new Object();
			// Deliver a dummy object as a requester to make sure that the classloader can be disposed. Once the program exits this scope, it can be released.
			final CloseableURLClassLoader classLoader = (CloseableURLClassLoader) this.getClassLoader(project, dummyObject);
			final ClassAndMethodContainer classAndMethodContainer = new ClassAndMethodContainer(classLoader);

			final List<PluginContainer> readers = new ArrayList<PluginContainer>();
			final List<PluginContainer> filters = new ArrayList<PluginContainer>();
			final List<RepositoryContainer> repositories = new ArrayList<RepositoryContainer>();

			// Update the components using all available dependencies
			final Collection<String> libs = this.listAllLibraries(project);
			for (final String lib : libs) {
				this.initializeAvailableComponents(readers, filters, repositories, this.getURL(lib, project), classLoader, classAndMethodContainer);
			}
			// Update the components using the Kieker library.
			this.initializeAvailableComponents(readers, filters, repositories, this.getKiekerURL(), classLoader, classAndMethodContainer);

			// Assemble the final object containing the available components - but make sure that the lists cannot be modified.
			final ComponentListContainer componentList = new ComponentListContainer(Collections.unmodifiableList(readers), Collections.unmodifiableList(filters),
					Collections.unmodifiableList(repositories));
			this.availableComponents.put(project, componentList);

			classLoader.close();
		} catch (final ProjectNotExistingException ex) {
			// Technically this should not happen. Log but ignore this exception.
			FSProjectDAOImpl.LOG.warn("An error occured while initializing the project '" + project + "'.", ex);
		}
	}

	/**
	 * This method initializes the available components for the given library and the given project. It runs through the library and extracts all available readers,
	 * filters and repositories.
	 * 
	 * @param readers
	 *            The list of readers which will be filled during the extraction process.
	 * @param filters
	 *            The list of filters which will be filled during the extraction process.
	 * @param repositories
	 *            The list of repositories which will be filled during the extraction process.
	 * @param lib
	 *            The url to the library which will be used for the extraction.
	 * @param classLoader
	 *            The classloader which will be used to load the classes.
	 * @param classAndMethodContainer
	 *            The object containing classes and methods for the extraction process.
	 * @throws IOException
	 *             If something went really wrong during the process.
	 */
	private void initializeAvailableComponents(final List<PluginContainer> readers, final List<PluginContainer> filters,
			final List<RepositoryContainer> repositories, final URL lib, final ClassLoader classLoader, final ClassAndMethodContainer classAndMethodContainer) throws
			IOException {
		// FInd the available classes within the library
		final List<Class<AbstractRepository>> repositoryClasses = this.pluginFinder.getAllRepositoriesWithinJar(lib, classLoader, classAndMethodContainer);
		final List<Class<AbstractPlugin>> pluginClasses = this.pluginFinder.getAllPluginsWithinJar(lib, classLoader, classAndMethodContainer);

		// Convert the repositories into model instances
		for (final Class<AbstractRepository> repository : repositoryClasses) {
			if (!(Modifier.isAbstract(repository.getModifiers()) || FSProjectDAOImpl.isProgrammaticOnly(repository, classAndMethodContainer))) {
				repositories.add(FSProjectDAOImpl.convertRepository(repository, classAndMethodContainer));
			}
		}

		// Convert the plugins into model instances
		for (final Class<AbstractPlugin> plugin : pluginClasses) {
			if (!(Modifier.isAbstract(plugin.getModifiers()) || FSProjectDAOImpl.isProgrammaticOnly(plugin, classAndMethodContainer))) {
				final PluginContainer pluginContainer = FSProjectDAOImpl.convertPlugin(plugin, classAndMethodContainer);
				if (pluginContainer.isReader()) {
					readers.add(pluginContainer);
				} else {
					filters.add(pluginContainer);
				}
			}
		}
	}

	/**
	 * This method converts the given class instance into a model instance.
	 * 
	 * @param clazz
	 *            The class to be converted.
	 * @param classAndMethodContainer
	 *            The container which will be used for the converting.
	 * @return An object containing the model instance.
	 */
	private static PluginContainer convertPlugin(final Class<AbstractPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		final Map<String, String> propertyDescriptions = new HashMap<String, String>();
		final Map<String, String> displayDescriptions = new HashMap<String, String>();
		boolean fullyInitialized;
		String description = "";
		String dependency = "";

		// Create a new instance for the model
		final MIPlugin plugin;
		if (classAndMethodContainer.getAbstractReaderPluginClass().isAssignableFrom(clazz)) {
			plugin = FSProjectDAOImpl.FACTORY.createReader();
		} else {
			plugin = FSProjectDAOImpl.FACTORY.createFilter();
		}
		plugin.setClassname(clazz.getName());
		plugin.setName(clazz.getSimpleName());

		// Those are the attributes which can go wrong because of the lazy class loading
		try {
			description = FSProjectDAOImpl.getDescription(clazz, classAndMethodContainer);
			dependency = FSProjectDAOImpl.getDependencies(clazz, classAndMethodContainer);

			FSProjectDAOImpl.fillProperties(clazz, plugin, classAndMethodContainer, propertyDescriptions);
			FSProjectDAOImpl.fillPorts(clazz, plugin, classAndMethodContainer);
			FSProjectDAOImpl.fillDisplays(clazz, plugin, classAndMethodContainer, displayDescriptions);

			fullyInitialized = true;
		} catch (final ComponentInitializationException ex) {
			// This exception can occur if (for example) a class is missing. We clear the component as far as possible - except for the description and the
			// dependencies. This is important
			plugin.getProperties().clear();
			plugin.getOutputPorts().clear();
			propertyDescriptions.clear();
			displayDescriptions.clear();
			if (plugin instanceof MIFilter) {
				((MIFilter) plugin).getInputPorts().clear();
			}
			// description = "";
			// dependency = "";

			fullyInitialized = false;

			FSProjectDAOImpl.LOG.warn("A component with the classname '" + clazz.getCanonicalName() + "' could not be initialized.", ex);
		}

		// Assemble the resulting instance with everything we could extract and return it
		return new PluginContainer(plugin, description, dependency, fullyInitialized, Collections.unmodifiableMap(propertyDescriptions),
				Collections.unmodifiableMap(displayDescriptions));
	}

	/**
	 * Tells whether the given class is "programmaticOnly". If there is no annotation which could tell the state of the flag, true will be returned.
	 * 
	 * @param clazz
	 *            The class of the plugin or repository.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection access.
	 * @return The state of the programmaticOnly flag of the plugin or repository.
	 */
	private static boolean isProgrammaticOnly(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		boolean result;
		try {
			// Get the two potential annotations
			final Annotation annotationPlugin = clazz.getAnnotation(classAndMethodContainer.getPluginAnnotationClass());
			final Annotation annotationRepository = clazz.getAnnotation(classAndMethodContainer.getRepositoryAnnotationClass());

			final Method pluginProgOnlyMethod = classAndMethodContainer.getPluginProgrammaticOnlyMethod();
			final Method repoProgOnlyMethod = classAndMethodContainer.getRepositoryProgrammaticOnlyMethod();

			// Now check which one of them is available
			if (annotationPlugin == null) {
				if (annotationRepository == null) {
					// None. Say it is programmatic only.
					result = true;
				} else {
					result = (Boolean) repoProgOnlyMethod.invoke(annotationRepository);
				}
			} else {
				result = (Boolean) pluginProgOnlyMethod.invoke(annotationPlugin);
			}

		} catch (final IllegalAccessException ex) {
			result = true;
		} catch (final IllegalArgumentException ex) {
			result = true;
		} catch (final InvocationTargetException ex) {
			result = true;
		} catch (final NoClassDefFoundError ex) {
			result = true;
		}

		return result;
	}

	/**
	 * This method converts the given class instance into a model instance.
	 * 
	 * @param clazz
	 *            The class to be converted.
	 * @param classAndMethodContainer
	 *            The container which will be used for the converting.
	 * @return An object containing the model instance.
	 */
	private static RepositoryContainer convertRepository(final Class<AbstractRepository> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		final Map<String, String> propertyDescriptions = new HashMap<String, String>();
		boolean fullyInitialized;
		String description = "";
		String dependency = "";

		// Create a new instance for the model
		final MIRepository repository = FSProjectDAOImpl.FACTORY.createRepository();

		repository.setClassname(clazz.getName());
		repository.setName(clazz.getSimpleName());

		// Those are the attributes which can go wrong because of the lazy class loading
		try {
			description = FSProjectDAOImpl.getDescription(clazz, classAndMethodContainer);
			dependency = FSProjectDAOImpl.getDependencies(clazz, classAndMethodContainer);
			FSProjectDAOImpl.fillProperties(clazz, repository, classAndMethodContainer, propertyDescriptions);

			fullyInitialized = true;
		} catch (final ComponentInitializationException ex) {
			// This exception can occur if (for example) a class is missing. We clear the component as far as possible - except for the description and the
			// dependencies. This is important
			repository.getProperties().clear();

			fullyInitialized = false;

			FSProjectDAOImpl.LOG.warn("A component with the classname '" + clazz.getCanonicalName() + "' could not be initialized.", ex);
		}

		// Assemble the resulting instance with everything we could extract and return it
		return new RepositoryContainer(repository, description, dependency, fullyInitialized, Collections.unmodifiableMap(propertyDescriptions));
	}

	/**
	 * This method can be used to get the description of an {@link AbstractPlugin}- or an {@link AbstractRepository}-class. The description is read via the
	 * annotation using the java reflection API.
	 * 
	 * @param clazz
	 *            The class whose description should be extracted.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @return The description for the class.
	 * @throws ComponentInitializationException
	 *             If the description could not be loaded.
	 */
	private static String getDescription(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) throws ComponentInitializationException {
		try {
			final String description;
			// Get the two potential annotations
			final Annotation annotationPlugin = clazz.getAnnotation(classAndMethodContainer.getPluginAnnotationClass());
			final Annotation annotationRepository = clazz.getAnnotation(classAndMethodContainer.getRepositoryAnnotationClass());

			final Method pluginDescrMethod = classAndMethodContainer.getPluginDescriptionMethod();
			final Method repoDescrMethod = classAndMethodContainer.getRepositoryDescriptionMethod();

			// Now check which one of them is available
			if ((annotationPlugin == null) || ((String) pluginDescrMethod.invoke(annotationPlugin)).isEmpty()) {
				if ((annotationRepository == null) || ((String) repoDescrMethod.invoke(annotationRepository)).isEmpty()) {
					// None. Deliver a human readable substitute.
					description = "No description available";
				} else {
					description = (String) repoDescrMethod.invoke(annotationRepository);
				}
			} else {
				description = (String) pluginDescrMethod.invoke(annotationPlugin);
			}

			return description;
		} catch (final IllegalAccessException ex) {
			throw new ComponentInitializationException("Could not load the description of the class.", ex);
		} catch (final IllegalArgumentException ex) {
			throw new ComponentInitializationException("Could not load the description of the class.", ex);
		} catch (final InvocationTargetException ex) {
			throw new ComponentInitializationException("Could not load the description of the class.", ex);
		} catch (final NoClassDefFoundError ex) {
			throw new ComponentInitializationException("Could not load the description of the class.", ex);
		}
	}

	/**
	 * Delivers the content of the dependencies-field of the given component (plugin or repository).
	 * 
	 * @param clazz
	 *            The class whose dependencies should be extracted.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @return The dependencies for the class.
	 * @throws ComponentInitializationException
	 *             If the dependencies could not be loaded.
	 */
	private static String getDependencies(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) throws ComponentInitializationException {
		try {
			final String description;

			// Get the two potential annotations
			final Annotation annotationPlugin = clazz.getAnnotation(classAndMethodContainer.getPluginAnnotationClass());
			final Annotation annotationRepository = clazz.getAnnotation(classAndMethodContainer.getRepositoryAnnotationClass());

			final Method pluginDepMethod = classAndMethodContainer.getPluginDependenciesMethod();
			final Method repoDepMethod = classAndMethodContainer.getRepositoryDependenciesMethod();

			// Now check which one of them is available
			if (annotationPlugin == null) {
				if (annotationRepository == null) {
					// None. Deliver a human readable substitute.
					description = "No dependency information available";
				} else {
					description = (String) repoDepMethod.invoke(annotationRepository);
				}
			} else {
				description = (String) pluginDepMethod.invoke(annotationPlugin);
			}

			return description;
		} catch (final IllegalAccessException ex) {
			throw new ComponentInitializationException("Could not load the dependencies of the class.", ex);
		} catch (final IllegalArgumentException ex) {
			throw new ComponentInitializationException("Could not load the dependencies of the class.", ex);
		} catch (final InvocationTargetException ex) {
			throw new ComponentInitializationException("Could not load the dependencies of the class.", ex);
		} catch (final NoClassDefFoundError ex) {
			throw new ComponentInitializationException("Could not load the dependencies of the class.", ex);
		}
	}

	/**
	 * This method fills the properties of the given repository or plugin. In other words: It tries to extract the list of properties from the given class (using the
	 * annotations) and to convert them into model instances.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param component
	 *            The component to be filled.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @param propertyDescriptions
	 *            The map which will later contain the properties description.
	 * @throws ComponentInitializationException
	 *             If the properties could not be loaded.
	 */
	private static void fillProperties(final Class<?> clazz, final MIAnalysisComponent component, final ClassAndMethodContainer classAndMethodContainer,
			final Map<String, String> propertyDescriptions) throws ComponentInitializationException {
		try {
			// Get the default configuration and use it to initialize the model repository
			final List<Annotation> properties = FSProjectDAOImpl.getProperties(clazz, classAndMethodContainer);

			for (final Annotation property : properties) {
				final MIProperty mProperty = FSProjectDAOImpl.FACTORY.createProperty();

				mProperty.setName((String) classAndMethodContainer.getPropertyNameMethod().invoke(property));
				mProperty.setValue((String) classAndMethodContainer.getPropertyDefaultValueMethod().invoke(property));

				if (component instanceof MIPlugin) {
					((MIPlugin) component).getProperties().add(mProperty);
				} else {
					((MIRepository) component).getProperties().add(mProperty);
				}

				// Get the description as well
				final String description = (String) classAndMethodContainer.getPropertyDescriptionMethod().invoke(property);
				propertyDescriptions.put(mProperty.getName(), description);
			}
		} catch (final IllegalAccessException ex) {
			throw new ComponentInitializationException("Could not load the properties of the class.", ex);
		} catch (final IllegalArgumentException ex) {
			throw new ComponentInitializationException("Could not load the properties of the class.", ex);
		} catch (final InvocationTargetException ex) {
			throw new ComponentInitializationException("Could not load the properties of the class.", ex);
		} catch (final NoClassDefFoundError ex) {
			throw new ComponentInitializationException("Could not load the properties of the class.", ex);
		}
	}

	/**
	 * Delivers the properties of the given class, using the annotations and the java reflection API.
	 * 
	 * @param clazz
	 *            The class whose properties will be delivered.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @return A list of properties. It contains instances of {@link Property} - but those from the project class loader.
	 * @throws InvocationTargetException
	 *             If something went wrong during the reflection invocation.
	 * @throws IllegalAccessException
	 *             If something went wrong during the reflection invocation.
	 * @throws IllegalArgumentException
	 *             If something went wrong during the reflection invocation.
	 */
	private static List<Annotation> getProperties(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {
		// Get the two potential annotations
		final Annotation annotationPlugin = clazz.getAnnotation(classAndMethodContainer.getPluginAnnotationClass());
		final Annotation annotationRepository = clazz.getAnnotation(classAndMethodContainer.getRepositoryAnnotationClass());

		final Annotation[] properties;

		// Now check which one of them is available
		if (annotationPlugin == null) {
			if (annotationRepository == null) {
				// None.
				properties = new Property[0];
			} else {
				properties = (Annotation[]) classAndMethodContainer.getRepositoryConfigurationMethod().invoke(annotationRepository);
			}
		} else {
			properties = (Annotation[]) classAndMethodContainer.getPluginConfigurationMethod().invoke(annotationPlugin);
		}

		return Arrays.asList(properties);
	}

	/**
	 * This method fills the displays of the given plugin. In other words: It tries to extract the list of displays from the given class (using the annotations) and
	 * to convert them into model instances.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @param displayDescriptions
	 *            The map which will later contain the display description.
	 * @throws ComponentInitializationException
	 *             If the properties could not be loaded.
	 */
	private static void fillDisplays(final Class<AbstractPlugin> clazz, final MIPlugin plugin, final ClassAndMethodContainer classAndMethodContainer,
			final Map<String, String> displayDescriptions) throws ComponentInitializationException {
		try {
			// Get the displays and convert them into model instances
			final List<Annotation> displays = FSProjectDAOImpl.getDisplays(clazz, classAndMethodContainer);

			for (final Annotation display : displays) {
				final MIDisplay mDisplay = FSProjectDAOImpl.FACTORY.createDisplay();
				mDisplay.setName((String) classAndMethodContainer.getDisplayNameMethod().invoke(display));
				plugin.getDisplays().add(mDisplay);

				// Get the description as well
				final String description = (String) classAndMethodContainer.getDisplayDescriptionMethod().invoke(display);
				displayDescriptions.put(mDisplay.getName(), description);
			}
		} catch (final IllegalAccessException ex) {
			throw new ComponentInitializationException("Could not load the displays of the class.", ex);
		} catch (final IllegalArgumentException ex) {
			throw new ComponentInitializationException("Could not load the displays of the class.", ex);
		} catch (final InvocationTargetException ex) {
			throw new ComponentInitializationException("Could not load the displays of the class.", ex);
		} catch (final NoClassDefFoundError ex) {
			throw new ComponentInitializationException("Could not load the displays of the class.", ex);
		}
	}

	/**
	 * This method fills the ports of the given plugin. In other words: It tries to extract the list of ports from the given class (using the annotations) and
	 * to convert them into model instances.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @throws ComponentInitializationException
	 *             If the properties could not be loaded.
	 */
	private static void fillPorts(final Class<AbstractPlugin> clazz, final MIPlugin plugin, final ClassAndMethodContainer classAndMethodContainer) throws
			ComponentInitializationException {
		try {
			// Get the ports
			final List<Annotation> inputPorts = FSProjectDAOImpl.getInputPorts(clazz, classAndMethodContainer);
			final List<Annotation> outputPorts = FSProjectDAOImpl.getOutputPorts(clazz, classAndMethodContainer);
			final List<Annotation> repositoryPorts = FSProjectDAOImpl.getRepositoryPorts(clazz, classAndMethodContainer);

			// Add input ports
			if (plugin instanceof MIFilter) {
				for (final Annotation inputPort : inputPorts) {
					final MIInputPort mInputPort = FSProjectDAOImpl.FACTORY.createInputPort();
					mInputPort.setName((String) classAndMethodContainer.getInputPortNameMethod().invoke(inputPort));
					mInputPort.setParent((MIFilter) plugin);
				}
			}

			// Add output ports.
			for (final Annotation outputPort : outputPorts) {
				final MIOutputPort mOutputPort = FSProjectDAOImpl.FACTORY.createOutputPort();
				mOutputPort.setName((String) classAndMethodContainer.getOutputPortNameMethod().invoke(outputPort));
				mOutputPort.setParent(plugin);
			}

			// Add repository ports.
			for (final Annotation repositoryPort : repositoryPorts) {
				final MIRepositoryConnector mConnector = FSProjectDAOImpl.FACTORY.createRepositoryConnector();
				mConnector.setName((String) classAndMethodContainer.getRepositoryPortNameMethod().invoke(repositoryPort));
				plugin.getRepositories().add(mConnector);
			}
		} catch (final IllegalAccessException ex) {
			throw new ComponentInitializationException("Could not load the ports of the class.", ex);
		} catch (final IllegalArgumentException ex) {
			throw new ComponentInitializationException("Could not load the ports of the class.", ex);
		} catch (final InvocationTargetException ex) {
			throw new ComponentInitializationException("Could not load the ports of the class.", ex);
		} catch (final NoClassDefFoundError ex) {
			throw new ComponentInitializationException("Could not load the ports of the class.", ex);
		}
	}

	/**
	 * Searches for input ports within the given class and returns them.
	 * 
	 * @param clazz
	 *            The class to be analyzed.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @return A list containing the available input ports.
	 */
	private static List<Annotation> getInputPorts(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		final List<Annotation> result = new ArrayList<Annotation>();

		for (final Method method : clazz.getMethods()) {
			// Get the potential annotation
			final Annotation annotationPort = method.getAnnotation(classAndMethodContainer.getInputPortAnnotationClass());
			// Now check whether it is available
			if (annotationPort != null) {
				result.add(annotationPort);
			}
		}

		return result;
	}

	/**
	 * Searches for output ports within the given class and returns them.
	 * 
	 * @param clazz
	 *            The class to be analyzed.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @return A list containing the available output ports.
	 * @throws InvocationTargetException
	 *             If something went wrong during the reflection invocation.
	 * @throws IllegalAccessException
	 *             If something went wrong during the reflection invocation.
	 * @throws IllegalArgumentException
	 *             If something went wrong during the reflection invocation.
	 */
	private static List<Annotation> getOutputPorts(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {
		final List<Annotation> result = new ArrayList<Annotation>();

		// Get the potential annotation
		final Annotation annotationPlugin = clazz.getAnnotation(classAndMethodContainer.getPluginAnnotationClass());

		// Now check whether it is available
		if (annotationPlugin != null) {
			for (final Annotation oPort : (Annotation[]) classAndMethodContainer.getPluginOutputPortsMethod().invoke(annotationPlugin)) {
				result.add(oPort);
			}
		}

		return result;
	}

	/**
	 * Searches for repository ports within the given class and returns them.
	 * 
	 * @param clazz
	 *            The class to be analyzed.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @return A list containing the available repository ports.
	 * @throws InvocationTargetException
	 *             If something went wrong during the reflection invocation.
	 * @throws IllegalAccessException
	 *             If something went wrong during the reflection invocation.
	 * @throws IllegalArgumentException
	 *             If something went wrong during the reflection invocation.
	 */
	private static List<Annotation> getRepositoryPorts(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {
		final List<Annotation> result = new ArrayList<Annotation>();

		// Get the potential annotation
		final Annotation annotationPlugin = clazz.getAnnotation(classAndMethodContainer.getPluginAnnotationClass());

		// Now check whether it is available
		if (annotationPlugin != null) {
			for (final Annotation rPort : (Annotation[]) classAndMethodContainer.getPluginRepositoryPortsMethod().invoke(annotationPlugin)) {
				result.add(rPort);
			}
		}

		return result;
	}

	/**
	 * Searches for displays within the given class and returns them.
	 * 
	 * @param clazz
	 *            The class to be analyzed.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection invocation.
	 * @return A list containing the available displays.
	 */
	private static List<Annotation> getDisplays(final Class<?> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		final List<Annotation> result = new ArrayList<Annotation>();

		for (final Method method : clazz.getMethods()) {
			// Get the potential annotation
			final Annotation display = method.getAnnotation(classAndMethodContainer.getDisplayAnnotationClass());
			// Now check whether it is available
			if (display != null) {
				result.add(display);
			}
		}

		return result;
	}

	/**
	 * Checks whether the given directory exists and creates it if necessary.
	 * 
	 * @param dirName
	 *            The name of the directory.
	 * @throws IOException
	 *             If the directory could not be created.
	 */
	private void checkDir(final String dirName) throws IOException {
		final File dir = new File(dirName);
		if (!dir.exists() && !dir.mkdir()) {
			// Try to create the directory
			throw new IOException("Could not create directory '" + dirName + "'.");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#addProject(java.lang.String)
	 */
	@Override
	@PreAuthorize("hasAnyRole('User', 'Administrator')")
	public void addProject(final String projectName, final String username) throws ProjectAlreadyExistingException, IOException {
		// Assemble all necessary paths and files for the given project
		final File projectDir = FSProjectDAOImpl.assembleProjectDir(projectName);
		final File projectFile = FSProjectDAOImpl.assembleKaxFile(projectName);
		final File metaFile = FSProjectDAOImpl.assembleMetaFile(projectName);
		final File libDir = this.assembleLibDir(projectName);

		// We need an "empty" project in order to save it.
		final MIProject emptyProject = FSProjectDAOImpl.FACTORY.createProject();

		// Make sure that the project doesn't exist already
		if (projectDir.exists()) {
			throw new ProjectAlreadyExistingException("The project with the name '" + projectName + "' exists already.");
		}

		try {
			// Try to create the directories and files
			if (projectDir.mkdir() && libDir.mkdir() && metaFile.createNewFile()) {
				// Try to save the file
				AnalysisController.saveToFile(projectFile, emptyProject);
				this.initializeAvailableComponents(projectName);

				// Store the initial meta data
				final Properties properties = new Properties();
				properties.put("owner", username);

				FileOutputStream out = null;
				try {
					out = new FileOutputStream(metaFile);
					properties.store(out, "");
				} finally {
					if (out != null) {
						out.close();
					}
				}
			} else {
				// The directories could not be created
				throw new IOException("Project-Directories could not be created.");
			}
		} catch (final IOException ex) {
			// Something went wrong. Remove the directories and files!
			final boolean libDirDeleted = libDir.delete();
			// Keep in mind that the potential remains of the file have to be deleted before the directory.
			final boolean projectFileDeleted = projectFile.delete();
			final boolean projectDeleted = projectDir.delete();

			// The following part is only necessary to calm FindBugs...
			@SuppressWarnings("unused")
			final boolean deleteResults = libDirDeleted && projectFileDeleted && projectDeleted;

			// Rethrow the exception in order to inform the caller of this method
			throw new IOException("An IO Exception occured while saving the project.", ex);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#copyProject(java.lang.String, java.lang.String)
	 */
	@Override
	@PreAuthorize("hasAnyRole('User', 'Administrator')")
	public void copyProject(final String originalProjectName, final String newProjectName) throws ProjectNotExistingException, ProjectAlreadyExistingException,
			IOException {
		// Get the necessary paths
		final File dstProjDir = FSProjectDAOImpl.assembleProjectDir(newProjectName);
		final File srcLibDir = this.assembleLibDir(originalProjectName);
		final File dstLibDir = this.assembleLibDir(newProjectName);

		// Check whether the project exists already!
		if (this.projectExists(newProjectName)) {
			throw new ProjectAlreadyExistingException("A project with the name '" + newProjectName + "' exists already.");
		}

		final File srcKaxFile = FSProjectDAOImpl.assembleKaxFile(originalProjectName);
		final File dstKaxFile = FSProjectDAOImpl.assembleKaxFile(newProjectName);

		try {
			if (dstProjDir.mkdir() && dstLibDir.mkdir()) {
				// Copy the kax file
				FileCopyUtils.copy(srcKaxFile, dstKaxFile);

				// Copy the libs
				for (final File lib : srcLibDir.listFiles()) {
					final File dstLibFile = new File(dstLibDir, lib.getName());
					FileCopyUtils.copy(lib, dstLibFile);
				}
			} else {
				throw new IOException("Could not create directories.");
			}
		} catch (final IOException ex) {
			// Something went wrong. Remove the directories and files!
			final boolean libDirDeleted = dstLibDir.delete();
			// Keep in mind that the potential remains of the file have to be deleted before the directory.
			final boolean projectFileDeleted = dstKaxFile.delete();
			final boolean projectDeleted = dstProjDir.delete();

			// The following part is only necessary to calm FindBugs...
			@SuppressWarnings("unused")
			final boolean deleteResults = libDirDeleted && projectFileDeleted && projectDeleted;

			// Rethrow the exception in order to inform the caller of this method
			throw new IOException("An IO Exception occured while copying the project.", ex);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#openProject(java.lang.String)
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public MIProject openProject(final String projectName) throws ProjectNotExistingException, IOException {
		if (projectName == null) {
			throw new IOException("Project is null");
		}

		try {
			// Load the project
			return AnalysisController.loadFromFile(FSProjectDAOImpl.assembleKaxFile(projectName).getAbsoluteFile());
		} catch (final FileNotFoundException ex) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.", ex);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#openProject(java.lang.String, kieker.webgui.common.ClassAndMethodContainer)
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public Object openProject(final String projectName, final ClassAndMethodContainer classAndMethodContainer) throws ProjectNotExistingException, IOException {
		if (projectName == null) {
			throw new IOException("Project is null");
		}

		// Load the project
		final Object project = ClassAndMethodContainer.invokeClassMethod(classAndMethodContainer.getAnalysisControllerLoadFromFile(), null,
				FSProjectDAOImpl.assembleKaxFile(projectName));

		if (project == null) {
			throw new IOException("Project could not be loaded.");
		} else {
			return project;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#saveProject(java.lang.String, kieker.analysis.model.analysisMetaModel.MIProject, long, boolean)
	 */
	@Override
	@PreAuthorize("hasAnyRole('User', 'Administrator')")
	public void saveProject(final String projectName, final MIProject project, final long timeStamp, final boolean overwriteNewerProject) throws
			ProjectNotExistingException, IOException, NewerProjectException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		// Check for a newer version first
		final long currTimeStamp = this.getCurrTimeStamp(projectName);
		if (!overwriteNewerProject && (currTimeStamp > timeStamp)) {
			throw new NewerProjectException("The project with the name '" + projectName + "' has a newer version on the FS.");
		}

		// Try to save it.
		AnalysisController.saveToFile(FSProjectDAOImpl.assembleKaxFile(projectName), project);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#getCurrTimeStamp(java.lang.String)
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public long getCurrTimeStamp(final String projectName) throws ProjectNotExistingException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}
		return FSProjectDAOImpl.assembleKaxFile(projectName).lastModified();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#uploadLibrary(org.primefaces.model.UploadedFile, java.lang.String)
	 */
	@Override
	@PreAuthorize("hasAnyRole('User', 'Administrator')")
	public void uploadLibrary(final UploadedFile file, final String projectName) throws ProjectNotExistingException, IOException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		// Prepare the files
		final File libDir = this.assembleLibDir(projectName);
		final File dstFile = new File(libDir, file.getFileName());

		// Get the streams
		final BufferedInputStream in = new BufferedInputStream(file.getInputstream());
		final BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(dstFile));

		// Copy the data - the streams will be closed in this method.
		FileCopyUtils.copy(in, out);

		// Reload the available components
		this.initializeAvailableComponents(projectName);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#getClassLoader(java.lang.String)
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public ClassLoader getClassLoader(final String projectName, final Object requester) throws ProjectNotExistingException, IOException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		// Create a new temporary directory
		final File tempDir = this.createTemporaryDirectory();
		final List<URL> libs = new ArrayList<URL>();

		// Copy and collect all libraries of the project
		final File libDir = new File(FSProjectDAOImpl.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSProjectDAOImpl.LIB_DIRECTORY);
		final File[] files = libDir.listFiles();
		if (files != null) {
			for (final File file : files) {
				if (file.getName().endsWith("." + FSProjectDAOImpl.LIB_EXTENSION)) {
					final File newLibFile = new File(tempDir, file.getName());
					FileCopyUtils.copy(file, newLibFile);
					try {
						libs.add(newLibFile.toURI().toURL());
					} catch (final MalformedURLException ex) {
						ex.printStackTrace();
					}
				}
			}
		}

		// Add the kieker lib!
		libs.add(this.getKiekerURL());

		// Now assemble the URL class loader
		final PrivilegedClassLoaderAction action = new PrivilegedClassLoaderAction(libs);
		final CloseableURLClassLoader classLoader = AccessController.doPrivileged(action);

		// Remember the requester
		final WeakReference<Object> ref = new WeakReference<Object>(requester);
		this.classLoaders.put(classLoader, ref);
		this.tempDirs.put(tempDir, ref);
		WeakReferenceMonitor.monitor(requester, this);

		return classLoader;
	}

	/**
	 * This method tries to create a new temporary directory within our temp directory.
	 * 
	 * @return A file object pointing to the new temporary directory.
	 */
	private File createTemporaryDirectory() {
		return Files.createTempDir();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#listAllLibraries(java.lang.String)
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public List<String> listAllLibraries(final String projectName) throws ProjectNotExistingException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		final List<String> result = new ArrayList<String>();

		// Run through the libs and put them into our list.
		final File[] files = this.assembleLibDir(projectName).listFiles();
		if (files != null) {
			for (final File file : files) {
				if (file.getName().endsWith("." + FSProjectDAOImpl.LIB_EXTENSION)) {
					result.add(file.getName());
				}
			}
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#listAllProjects()
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public Collection<String> listAllProjects() {
		final List<String> result = new ArrayList<String>();

		// Get all directories within our root-dir
		final File[] files = new File(FSProjectDAOImpl.ROOT_DIRECTORY).listFiles();
		for (final File file : files) {
			if (file.isDirectory()) {
				result.add(file.getName());
			}
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#getAvailableComponents(java.lang.String)
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public ComponentListContainer getAvailableComponents(final String project) {
		return this.availableComponents.get(project);
	}

	/**
	 * Checks whether a project with the name exists on the file system.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return true if and only if a directory with the name of the project exists in the root dir.
	 */
	private boolean projectExists(final String projectName) {
		return FSProjectDAOImpl.assembleKaxFile(projectName).exists();
	}

	/**
	 * Delivers the url of the given library to the given project.
	 * 
	 * @param lib
	 *            The library of the project.
	 * @param project
	 *            The project itself.
	 * @return An URL pointing to the URL.
	 * @throws MalformedURLException
	 *             If something went wrong.
	 */
	private URL getURL(final String lib, final String project) throws MalformedURLException {
		final File file = new File(FSProjectDAOImpl.ROOT_DIRECTORY + File.separator + project + File.separator + FSProjectDAOImpl.LIB_DIRECTORY + File.separator
				+ lib);
		return file.toURI().toURL();
	}

	/**
	 * Assembles the {@link File}-element pointing to the directory of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The directory of the project.
	 */
	private static File assembleProjectDir(final String projectName) {
		return new File(FSProjectDAOImpl.ROOT_DIRECTORY + File.separator + projectName);
	}

	/**
	 * Assembles the {@link File}-element pointing to the kax-file of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The kax-file of the project.
	 */
	private static File assembleKaxFile(final String projectName) {
		return new File(FSProjectDAOImpl.ROOT_DIRECTORY + File.separator + projectName + File.separator + projectName + "." + FSProjectDAOImpl.KAX_EXTENSION);
	}

	/**
	 * Assembles the {@link File}-element pointing to the meta file of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The meta file of the project.
	 */
	private static File assembleMetaFile(final String projectName) {
		return new File(FSProjectDAOImpl.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSProjectDAOImpl.META_FILE);
	}

	/**
	 * Assembles the {@link File}-element pointing to the library directory of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The library directory of the project.
	 */
	private File assembleLibDir(final String projectName) {
		return new File(FSProjectDAOImpl.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSProjectDAOImpl.LIB_DIRECTORY);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#deleteLibrary(java.lang.String, java.lang.String)
	 */
	@Override
	@PreAuthorize("hasAnyRole('User', 'Administrator')")
	public boolean deleteLibrary(final String projectName, final String libName) throws IOException {
		final File libDir = this.assembleLibDir(projectName);
		final File libFile = new File(libDir, libName);
		final boolean result = libFile.delete();
		// Reload the available components
		this.initializeAvailableComponents(projectName);

		return result;
	}

	/**
	 * Delivers an URL pointing to the kieker library within this application.
	 * 
	 * @return The kieker url.
	 */
	private URL getKiekerURL() {
		return Thread.currentThread().getContextClassLoader().getResource(FSProjectDAOImpl.KIEKER_LIB);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#getProjectFile(java.lang.String)
	 */
	@Override
	@PreAuthorize("isAuthenticated()")
	public File getProjectFile(final String projectName) {
		return FSProjectDAOImpl.assembleKaxFile(projectName);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.util.WeakReferenceMonitor.ReleaseListener#released()
	 */
	@Override
	public void released() {
		synchronized (this.classLoaders) {
			int open = 0;

			final List<CloseableURLClassLoader> toBeClosed = new ArrayList<CloseableURLClassLoader>();
			final List<File> toBeRemoved = new ArrayList<File>();

			// Run through the class loaders and check which of them can be closed
			for (final Entry<CloseableURLClassLoader, WeakReference<Object>> entry : this.classLoaders.entrySet()) {
				if (entry.getValue().get() == null) {
					toBeClosed.add(entry.getKey());
				} else {
					open++;
				}
			}
			for (final Entry<File, WeakReference<Object>> entry : this.tempDirs.entrySet()) {
				if (entry.getValue().get() == null) {
					toBeRemoved.add(entry.getKey());
				}
			}

			for (final CloseableURLClassLoader classLoader : toBeClosed) {
				try {
					classLoader.close();
					this.classLoaders.remove(classLoader);
					FSProjectDAOImpl.LOG.info("Closed classloader (" + classLoader + ")");
				} catch (final IOException ex) {
					FSProjectDAOImpl.LOG.error("Could not close classloader (" + classLoader + ")");
				}
			}
			for (final File tempDir : toBeRemoved) {
				final boolean result = FileSystemUtils.deleteRecursively(tempDir);
				if (result) {
					this.tempDirs.remove(tempDir);
					FSProjectDAOImpl.LOG.info("Removed temporary directory (" + tempDir + ")");
				} else {
					FSProjectDAOImpl.LOG.error("Could not remove temporary directory (" + tempDir + ")");
				}
			}
			FSProjectDAOImpl.LOG.info(open + " classloaders still open.");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IProjectDAO#getOwner(java.lang.String)
	 */
	@Override
	public String getOwner(final String projectName) {
		final File metaFile = FSProjectDAOImpl.assembleMetaFile(projectName);
		final Properties properties = new Properties();
		InputStream inputStream = null;
		try {
			inputStream = new FileInputStream(metaFile);
			properties.load(inputStream);
			return properties.getProperty("owner");
		} catch (final IOException ex) {
			FSProjectDAOImpl.LOG.warn("Could not open meta file.", ex);
		} finally {
			try {
				if (inputStream != null) {
					inputStream.close();
				}
			} catch (final IOException ex) {
				FSProjectDAOImpl.LOG.warn("Could not open meta file.", ex);
			}
		}

		return "N/A";
	}

	/**
	 * This helper class is responsible for creating a classloader as a privileged action. This is recommended due to the java security manager.
	 * 
	 * @author Nils Christian Ehmke
	 * @version 1.0
	 */
	private static class PrivilegedClassLoaderAction implements PrivilegedAction<CloseableURLClassLoader> {
		/**
		 * The list of libraries used to create the class loader.
		 */
		private final List<URL> libs;

		/**
		 * Creates a new instance of this class using the given parameters.
		 * 
		 * @param libs
		 *            The list of libraries used to create the class loader.
		 */
		public PrivilegedClassLoaderAction(final List<URL> libs) {
			this.libs = libs;
		}

		/**
		 * Runs the action.
		 * 
		 * @return The class loader.
		 */
		@Override
		public CloseableURLClassLoader run() {
			// We use "null" as an explicit parent to make sure that the class loader is completely independent from the web gui class loader.
			return new CloseableURLClassLoader(this.libs.toArray(new URL[this.libs.size()]), null);
		}
	}

}
